/*  OctoWS2811 BasicTest.ino - Basic RGB LED Test
    http://www.pjrc.com/teensy/td_libs_OctoWS2811.html
    Copyright (c) 2013 Paul Stoffregen, PJRC.COM, LLC

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
    THE SOFTWARE.

  Required Connections
  --------------------
    pin 2:  LED Strip #1    OctoWS2811 drives 8 LED Strips.
    pin 14: LED strip #2    All 8 are the same length.
    pin 7:  LED strip #3
    pin 8:  LED strip #4    A 100 ohm resistor should used
    pin 6:  LED strip #5    between each Teensy pin and the
    pin 20: LED strip #6    wire to the LED strip, to minimize
    pin 21: LED strip #7    high frequency ringining & noise.
    pin 5:  LED strip #8
    pin 15 & 16 - Connect together, but do not use
    pin 4 - Do not use
    pin 3 - Do not use as PWM.  Normal use is ok.

  This test is useful for checking if your LED strips work, and which
  color config (WS2811_RGB, WS2811_GRB, etc) they require.
*/

#include <OctoWS2811.h>

const int LED_COUNT = 180;
const int LED_BUFFER_SIZE = 240;

const uint_fast8_t COLOR_TEST_WIDTH = 16;
const uint_fast8_t COLOR_TEST_OFFSET = 24;

DMAMEM int displayMemory[LED_BUFFER_SIZE*6];
int drawingMemory[LED_BUFFER_SIZE*6];

const int config = WS2811_RGB | WS2811_800kHz;

OctoWS2811 leds(LED_BUFFER_SIZE, displayMemory, drawingMemory, config);

float phase = 0;

float hue = 0;   // 0..1

const uint_fast8_t BRIGHT = 0x1f;

void setup() {
  leds.begin();

  for (int i=0; i<LED_BUFFER_SIZE; i++) {
    leds.setPixel(i, 0x0);
  }

  leds.show();
}

inline float gamma( float v ) {
  return v;
  //return pow( v, 2.2f );
}

inline float lerp( float a, float b, float ratio ) {
  return a + (b-a)*ratio;
}

inline float getHueValue( float h, float offset ) {
  h = fmodf( h + offset, 1.0f );

  if( h < 0.16667f ) {
    return 1.0f;
  } else if( h < 0.33333f ) {
    return 1.0f - (h-0.16667f)*6.0f;
  } else if( h > 0.83333f ) {
    return 1.0f;
  } else if( h > 0.66667f ) {
    return (h-0.66667f)*6.0f;
  }

  return 0;
}

void loop() {
  float h = fmodf( hue, 1.0f );
  float red = getHueValue( h, 0.0f );
  float green = getHueValue( h, 1.0f - 0.33333f );
  float blue = getHueValue( h, 1.0f - 0.66667f );

  for( uint_fast8_t i=0; i<COLOR_TEST_WIDTH; i++ ) {
    float s = sin( phase - i*0.7f );
    float t = sin( s*2 * -cos(phase+i*0.3f) );

    float brt = (t+1.0f) * 0.5f;  // 0..1
    float colorAmt = min( brt*2.0f, 1.0f );
    float whiteAmt = max( 0, brt*2.0f - 1.0f );

    uint_fast8_t r = gamma(lerp(red,1.0f,whiteAmt)*colorAmt) * BRIGHT;
    uint_fast8_t g = gamma(lerp(green,1.0f,whiteAmt)*colorAmt) * BRIGHT;
    uint_fast8_t b = gamma(lerp(blue,1.0f,whiteAmt)*colorAmt) * BRIGHT;
    //uint_fast32_t color = (r<<16) | (g<<8) | b;

    setPixel_no_W( i, r, g, b );
    setPixel_colorMinusPureWhite( i+COLOR_TEST_OFFSET, r, g, b );
    setPixel_colorMinusWarmWhite( i+COLOR_TEST_OFFSET*2, r, g, b );
    setPixel_extra_W( i+COLOR_TEST_OFFSET*3, r, g, b );
  }

  leds.show();

  phase += 0.0075f;

  hue += random( 0, 9999 ) * (0.00033f/9999.0f);
  if( hue > 1.0f ) {
    hue -= 1.0f;
  }
}

void apply_led_W(int idx, uint_fast8_t w) {
  int colorIdx = (idx * 4) / 3;

  // Set the white
  int whiteIdx = colorIdx/4;  // round to nearest group of 3 LEDs
  whiteIdx = whiteIdx * 4 + 3;  // target the final 3 color components. Chipset order is: RGB,RGB,RGB,WWW

  int oldW = leds.getPixel(whiteIdx);
  int newW = 0;

  // Holy cats, the W component is bright
  //w = random(0,0xff);

  switch( colorIdx % 4 ) {
    case 0:  newW = (oldW&0x00ffff) | (w<<16); break;
    case 1:  newW = (oldW&0xff00ff) | (w<<8); break;
    case 2:  newW = (oldW&0xffff00) | w; break;
  }

  leds.setPixel( whiteIdx, newW );
}

void setPixel_no_W(int idx, uint_fast8_t r, uint_fast8_t g, uint_fast8_t b)
{
  int colorIdx = (idx * 4) / 3;
  leds.setPixel(colorIdx, (r<<16)|(g<<8)|b);
}

void setPixel_colorMinusPureWhite(int idx, uint_fast8_t r, uint_fast8_t g, uint_fast8_t b)
{
  int colorIdx = (idx * 4) / 3;

  uint_fast8_t w = min( min( r, g ), b );

  // Set the colorful RGB part
  int colorMinusWhite = ((r-w)<<16) | ((g-w)<<8) | (b-w);
  leds.setPixel(colorIdx, colorMinusWhite);

  apply_led_W(idx, w);
}

const uint_fast8_t WARM_R = 0xff;
const uint_fast8_t WARM_G = 0xeb;
const uint_fast8_t WARM_B = 0xd4;

void setPixel_colorMinusWarmWhite(int idx, uint_fast8_t r, uint_fast8_t g, uint_fast8_t b)
{
  int colorIdx = (idx * 4) / 3;

  float w_f = min( min( r*(1.0f/WARM_R), g*(1.0f/WARM_G) ), b*(1.0f/WARM_B) );
  w_f = min( 1.0f, w_f );

  int colorMinusWarmWhite =
    ((uint_fast8_t)(r - w_f*WARM_R) << 16) |
    ((uint_fast8_t)(g - w_f*WARM_G) << 8)  |
     (uint_fast8_t)(b - w_f*WARM_B);

  leds.setPixel(colorIdx, colorMinusWarmWhite);
  apply_led_W(idx, w_f * 0xff);
}

void setPixel_extra_W(int idx, uint_fast8_t r, uint_fast8_t g, uint_fast8_t b)
{
  int colorIdx = (idx * 4) / 3;

  uint_fast8_t w = min( min( r, g ), b );

  // Set the color
  leds.setPixel(colorIdx, (r<<16)|(g<<8)|(b));

  apply_led_W(idx, w);
}
